// DNA Sequenzierungsprogramm v2.0, 16/04/2001
// Jan Wassenberg
program dna_seq;

uses
	classes;

const
	dna_len    = 10*1000;	// Lnge des DNA-Strangs [Nukleotiden]
	enzyme_len = 4;			// Lnge eines Enzyms [Nukleotiden];

type
	// Enzymbeschreibung
	tenzyme = record
		str		: string[enzyme_len];	// wonach im DNA string gesucht wird 	; bei "a|tg" = "atg"
		cut_pos	: 2..enzyme_len;		// ab str[cut_pos] wird abgeschnitten	; bei "a|tg" = 2
		end_str	: string[enzyme_len-1];	// = copy(str, cut_pos, enzyme_len);	; bei "a|tg" = "tg"
		user_str: string[enzyme_len+1];	// String der Form "a|tg" fr Ausgaben	; bei "a|tg" = "a|tg"
	end; // of tenzyme

var
	dna, new_dna : string; // Origineller bzw. neu sequenzierter DNA-Strang
	enzyme1, enzyme2 : tenzyme; // Enzymdarstellungen [s. tenzyme oben]
	lists : array[1..2] of tstringlist; // 2 Listen von DNA-Fragmenten,
										// die von enzyme1 bzw. enzyme2 (s.o.) gespalten wurden

	frag : string; // Das aktuelle Fragment, das gerade bearbeitet wird
	list : 1..2; // Die Fragmentliste, aus der das passende Fragment zu frag kommen wird

	i : 1..dna_len; // Zhler zur Erzeugung der DNA

const
	bases : string[4] = 'actg'; // Nukleotiden

// Falls die aktuelle DNA-Enzymkombination ungnstig (s. "goto restart" unten) ist,
// fngt das Programm mit neu erzeugten Enzymen bei restart von Vorne an 
label
	restart;


//---------------------------------------------------------------------------------------------------------
// Erzeugt eine Enzymbeschreibung nach folgenden Regeln:
// - es gibt genau enzyme_len (<= 4) unterschiedliche Nukleotiden in der Enzymbeschreibung
// - die Trennstelle, ab der (einschliesslich) getrennt wird,
//   ist eine Position in dem Enzymbeschreibungsstring.
//   bei z.B. "a|tg" wre diese Position 2.
//                                                         
// Die Enzymbeschreibung enthlt verschiedene Darstellungsformen:
// - str ist die Nukleotidensequenz, die im DNA vorliegen muss, damit das Enyzm andocken kann
// - user_str enthlt das '|' Zeichen als Trennstelle und wird fr Ausgaben verwendet
// - end_str wird zur Prfung, ob 2 Enzyme gleich anfangende Fragmente erzeugen, verwendet
procedure make_enzyme(var enzyme : tenzyme);
var
	i : 1..enzyme_len;
	base_num : 1..4;
begin
	with enzyme do begin
		// str mit zuflligen Nukleotiden fllen
		setlength(str, enzyme_len);
		for i := 1 to enzyme_len do begin
			// Ein noch nicht verwendetes Nukleotid finden und in str[i] speichern
			repeat
				base_num := random(4)+1;
				str[i] := bases[base_num];
			until str[i] <> ' ';
			bases[base_num] := ' ';
		end; // of for

		// eine zufllige Trennstelle erzeugen
		cut_pos := random(enzyme_len-1)+2; // cut_pos liegt in {2..4} bei enzyme_len = 4

		// end_str ist das hintere Teil der Enzymbeschreibung (das nach der Trennstelle),
		// und damit auch die Anfangssequenz aller Fragmente, die von diesem Enzym abgespalten wurden
		end_str := copy(str, cut_pos, enzyme_len);

		// Die Enzymbeschreibungsform fr Benutzerausgaben initialisieren
		user_str := str;
		insert('|', user_str, cut_pos);
	end; // of with

	// bases wieder 'reparieren'
	bases := 'actg';
end;


//---------------------------------------------------------------------------------------------------------
// Teilt einen DNA-Strang mit dem vorgegebenen Enzym in eine Liste von Fragmenten auf
procedure split_dna(var list : tstringlist; dna : string; enzyme : tenzyme);
var
	cut_pos : integer;
begin
	list := tstringlist.create;

	// Erste Andockposition des Enzyms in dna ermitteln:
	// Suche nach der Nukleotidsequenz des Enzyms im DNA-String.
	// Falls keine vorhanden ist, wird der DNA-Strang als 1 grosses Fragment gespeichert
	cut_pos := pos(enzyme.str, dna);
	if cut_pos = 0 then cut_pos := maxint; // Sicherstellen, dass der gesamte Strang
										   // als Fragment kopiert wird

	// Spalte den DNA-Strang - dazu wird der Teil des DNA-Strangs bis zu Trennstelle
	// aus der DNA gelscht und der Fragmentliste zugefgt.
	// Dies wird solange durchgefhrt, bis es keine Andockstelle mehr gibt.
	repeat
		// Erstes Fragment der Fragmentliste hinzufgen...
		list.add(copy(dna, 1, cut_pos+enzyme.cut_pos-2));
		// und aus dem DNA-Strang lschen
		delete(dna, 1, cut_pos+enzyme.cut_pos-2);

		// Nchste Andockposition ermitteln
		cut_pos := pos(enzyme.str, dna);
	until cut_pos = 0; // wiederholen, bis es keine Andockstelle mehr gibt

	// Den Rest des DNA-Strangs als letztes Fragment speichern;
	// die Liste wird alphabetisch sortiert, um 'real-world' conditions zu simulieren
	// (die Fragmente sind nicht mehr in der Reihenfolge, in der sie abgespalten wurden).
	list.add(dna);
	list.sort;

	// Hinweis: es wird vorausgesetzt, dass die Fragmente eine eindeutige,
	// bekannte Orientierung von links nach rechts haben: das einzige Fragment,
	// das nicht mit end_str anfngt, ist das Anfangsfragment.
	// Die Sequenzierhardware muss 3' und 5' Richtung unterscheiden knnen.
end;


// Den Anfangsfragment in list (das durch enzyme gespalten wurde) suchen:
// falls es ein Fragment gibt, das nicht mit dem end_str von enzyme anfngt,
// ist dies das Anfangsfragment. Anderenfalls wird '' zurckgegeben.
function get_start_frag(enzyme : tenzyme; list : tstringlist) : string;
var
	i : integer;
begin
	// Alle Fragmente analysieren
	for i := 0 to list.count-1 do
		// Falls das aktuelle Fragment nicht mit end_str anfngt...
		if copy(list[i], 1, length(enzyme.end_str)) <> enzyme.end_str then begin
			// ist es das Anfangsfragment
			get_start_frag := list[i];
			exit;
		end; // of if

	// Kein Anfangsfragment gefunden; suche in der anderen Liste
	get_start_frag := '';
end;

//---------------------------------------------------------------------------------------------------------
// Diese andere Liste nach einem Fragment durchsuchen, das zum aktuellen Fragment passt.
// Dazu wird zunchst die Anzahl Fragmente in count gespeichert
// (wird spter bei der Eindeutigkeitsprfung bentigt).
// Dann werden alle Fragmente ermittelt, die gleich anfangen wie frag:
// wenn copy(temp, 1, length(frag)) = copy(frag, 1, length(temp)), fangen
// temp und frag gleich an.
// Jedes solche Fragment, das gefunden wird, wird in str gespeichert und aus der Liste entfernt.
// Anschliessend wird geprft, wieviele Fragmente aus der Liste entfernt wurden.
// Betrgt diese Anzahl nicht 1, wurden entweder keins oder zuviele passende Fragmente gefunden.
// In diesem Fall wird die Sequenzierung mit anderen Enzymen neu gestartet.
// Falls nur ein Fragment gefunden wurde, wird der gemeinsame Teil von den beiden passenden
// Fragmenten ermittelt, in substr gespeichert und dem rekombinierten DNA-Strang angehngt.
// Der berhang muss ermittelt werden: dazu wird das lngere Fragment in frag gespeichert;
// von diesem Fragment wird der gemeinsame Anfang gelscht.
// Damit bleibt nur der berhang, der als nchstes Anfangsfragment benutzt wird.
function match_frag : boolean;
var
	i, count : integer;
	temp : string;
	str, substr : string;
begin
	count := lists[list].count-1;

	// Passende Fragmente suchen
	for i := count downto 0 do begin
		temp := lists[list][i];
		// Falls die Fragmente frag und temp gleich anfangen,
		if copy(temp, 1, length(frag)) = copy(frag, 1, length(temp)) then begin
			// speichere temp in str und entferne es aus der Liste
			str := temp;
			lists[list].delete(i);
		end; // of if
	end; // of for

	// Falls entweder kein oder zuviele passende Fragmente gefunden wurden,
	// wird false zurckgegeben (die Sequenzierung wird mit einer neuen Enzymkombination neu gestartet) 
	if lists[list].count <> count then begin
		match_frag := false;
		exit;
	end; // of if

	// substr ist der gemeinsame Anfang von str und frag;
	// hnge dies dem neuen DNA-Strang hinten an
	substr := copy(str, 1, length(frag));
	new_dna := new_dna + substr;

	// Falls das zu frag passende Fragment lnger als frag ist, wird das nchste Fragment
	// aus der anderen Liste kommen; es wird sichergestellt, das frag das lngere
	// der beiden passenden Fragmente ist
	if length(str) > length(frag) then begin
		frag := str;
		list := 3-list;
	end; // of if

	// Der berhang ist das Stck des lngeren Fragments (s. Block oben), das nach substr kommt.
	delete(frag, 1, length(substr)); // frag = berhang

	// Erfolg melden
	match_frag := true;
end;

//---------------------------------------------------------------------------------------------------------
// Enzyme, originalen und neu sequenzierten DNA-Strang in einer Datei speichern
procedure dump_data;
var
	f : textfile;
begin
	// Keine Ausgabe, falls kein Kommandozeilenparameter bergeben wurde
	if paramcount = 0 then exit;

	assignfile(f, paramstr(1));
	rewrite(f);

	try
		writeln(f, 'enzyme1.user_str: "', enzyme1.user_str, '"; enzyme2.user_str: "', enzyme2.user_str, '"');
		writeln(f);
		writeln(f, 'dna:');
		writeln(f, '>', dna, '<');
		if new_dna = dna then writeln(f, ' =') else writeln(f, ' <>');
		writeln(f, 'new_dna:');
		writeln(f, '>', new_dna, '<');
	finally
		closefile(f);
	end; // of finally
end;

//---------------------------------------------------------------------------------------------------------
begin
	randomize;

	// DNA-Strang mit zuflligen Nukleotiden fllen
	setlength(dna, dna_len);

	for i := 1 to dna_len do
		dna[i] := bases[random(4)+1];

restart:
	new_dna := '';
	list := 2;

	// Enzyme erzeugen
	make_enzyme(enzyme1);
	make_enzyme(enzyme2);

	// Den DNA-Strang durch enzyme1 bzw. enzyme2 spalten
	split_dna(lists[1], dna, enzyme1);
	split_dna(lists[2], dna, enzyme2);

	// Das Anfangsfragment ermitteln:
	frag := get_start_frag(enzyme1, lists[1]);
	if frag = '' then begin // Anfangsfragment konnte aus Liste 1 nicht ermittelt werden
		// Suche in der zweiten Fragmentliste
		frag := get_start_frag(enzyme2, lists[2]);
		if frag = '' then goto restart; // Falls das Anfangsfragment nicht differenzierbar ist,
										// starte die Sequenzierung mit anderen Enzymen neu
		list := 1; // Das Anfangsfragment kam aus Liste 2 - das erste passende Fragment wird aus Liste 1 kommen
	end;

	// Ein passendes Fragment zu frag finden und bearbeiten,
	// bis es keine Fragmente mehr gibt, oder keins gefunden wurde
	repeat
		if not match_frag then goto restart; // Bei Fehler oder Uneindeutigkeit mit anderen Enzymen neu starten
	until lists[list].count = 0;

	dump_data;

	// Sicherstellen, dass das sequenzierte DNA dem originalen DNA-Strang entspricht
	assert(dna = new_dna, 'Fehler: der alte DNA-Strang entspricht nicht dem neu sequenzierten');
end.
